#include "monsters.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "../common/sources/Map.h"
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "player.h"
#include "walls.h"
#include "interface.h"
#include "game.h"
#include "sound.h"

CMonsters    monsters;

//---------------------------------------------------------------------------------------------
CMonsters::CMonsters()
{

}

//---------------------------------------------------------------------------------------------
void CMonsters::readMonstersDB()
{
    QDomDocument doc;

    QFile f("databases/monsters.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    monsterNode = root.firstChild();

    while(!monsterNode.isNull())
    {
        if (monsterNode.isElement())
        {
            QDomElement monster = monsterNode.toElement();
            std::vector<CParamType> paramsTypesList;

            if (monster.tagName() == "monster")
            {
                QDomElement monsterInfo = monster.firstChild().toElement();
                SMonsterData    monsterData;

                while(!monsterInfo.isNull())
                {
                    if (monsterInfo.tagName() == "img_front")
                    {
                        monsterData.frontImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_side")
                    {
                        monsterData.sideImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_back")
                    {
                        monsterData.backImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_attack")
                    {
                        monsterData.attackImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "time_between_atk")
                    {
                        monsterData.timeBetweenAttack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_display_time")
                    {
                        monsterData.attackDisplayTime = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_sound")
                    {
                        monsterData.attackSound = QString("sound/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "defense")
                    {
                        monsterData.defense = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "health")
                    {
                        monsterData.baseHealth = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "attack")
                    {
                        monsterData.attack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "dexterity")
                    {
                        monsterData.dexterity = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "param")
                    {
                        QString paramType = monsterInfo.attribute("type");

                        CParamType  newParamType;

                        if (paramType == "int")
                        {
                            newParamType.mType = eParamInt;
                            newParamType.mName = monsterInfo.text();
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "enum")
                        {
                            newParamType.mType = eParamEnum;
                            newParamType.mName = monsterInfo.text();
                            newParamType.mValues = monsterInfo.attribute("values").split(";");
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "stack")
                        {
                            newParamType.mType = eParamStack;
                            paramsTypesList.push_back(newParamType);
                        }
                    }

                    monsterInfo = monsterInfo.nextSibling().toElement();
                }

                CMap::mMonstersParams.push_back(paramsTypesList);
                monstersDatas.push_back(monsterData);
            }
        }
        monsterNode = monsterNode.nextSibling();
    }
}

//---------------------------------------------------------------------------------------------
int CMonsters::getStartingLife(int type)
{
    int baseHealth = monstersDatas[type].baseHealth;

    return baseHealth * game.levelMultiplier() + RANDOM((baseHealth / 4) + 1);
}

//---------------------------------------------------------------------------------------------
void CMonsters::init()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2  group;
        int nbMonsters = mapGroup.getIntParam("nbMonsters");

        group.dir = (EMonsterDir)mapGroup.getEnumParam("Dir");

        for (int j = 0; j < 4; j++)
        {
            group.monsters[j].dir = group.dir;
            group.monsters[j].pos = eMonsterPosNone;
            group.monsters[j].life = 0;
        }

        if (nbMonsters == 1)
        {
            group.monsters[0].pos = eMonsterPosCenter;
            group.monsters[0].life = getStartingLife(mapGroup.getType());
        }
        else
        {
            // monster positions for a given direction
            static const EMonsterPos startingPos[4][4] =
            {
                {eMonsterPosNorthWest, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosSouthEast},   // up
                {eMonsterPosSouthEast, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosNorthWest},   // down
                {eMonsterPosNorthWest, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosSouthEast},   // left
                {eMonsterPosSouthEast, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosNorthWest}    // right
            };

            for (int j = 0; j < 4; j++)
            {
                if (j < nbMonsters)
                {
                    group.monsters[j].pos = startingPos[group.dir][j];
                    group.monsters[j].life = getStartingLife(mapGroup.getType());
                }
            }
        }

        monsterGroups.push_back(group);
    }
}

//---------------------------------------------------------------------------------------------
CVec2 CMonsters::getMonsterPos(float x, float y)
{
    CVec2   pos;
    pos.x = 250 * (x - 4.5) / (8.5 - y) + 112;
    pos.y = 310 / (8.5 - 0.9 * y) + 66;
    return pos;
}

//---------------------------------------------------------------------------------------------
void CMonsters::draw(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    if (tablePos.y == WALL_TABLE_HEIGHT - 1)
        return;

    int groupIndex = map.findMonsterGroupIndex(mapPos);

    if (groupIndex != -1)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[groupIndex];
        CMonsterGroup2& group = monsterGroups[groupIndex];

        // monster positions depending on the player direction
        static const float monsterPosDir[4][5][2] =
        {
            //   NW          NE          SW          SE           C
            {{0.0, 0.0}, {1.0, 0.0}, {0.0, 1.0}, {1.0, 1.0},  {0.5, 0.5}},   // up
            {{1.0, 0.0}, {1.0, 1.0}, {0.0, 0.0}, {0.0, 1.0},  {0.5, 0.5}},   // left
            {{1.0, 1.0}, {0.0, 1.0}, {1.0, 0.0}, {0.0, 0.0},  {0.5, 0.5}},   // down
            {{0.0, 1.0}, {0.0, 0.0}, {1.0, 1.0}, {1.0, 0.0},  {0.5, 0.5}}    // right
        };

        // monster dir depending on the player direction
        static const EMonsterDir monsterDirDir[4][4] =
        {

            {eMonsterDirUp,    eMonsterDirDown,  eMonsterDirLeft,  eMonsterDirRight}, // up
            {eMonsterDirRight, eMonsterDirLeft,  eMonsterDirUp,    eMonsterDirDown},  // left
            {eMonsterDirDown,  eMonsterDirUp,    eMonsterDirRight, eMonsterDirLeft},  // down
            {eMonsterDirLeft,  eMonsterDirRight, eMonsterDirDown,  eMonsterDirUp}     // right
        };

        // we draw the back row monsters before the front ones
        for (int row = 0; row < 3; ++row)
        {
            for (int i = 0; i < 4; ++i)
            {
                CMonster&   monster = group.monsters[i];

                if (monster.pos != eMonsterPosNone)
                {
                    // relative direction
                    EMonsterDir tableDir = monsterDirDir[player.dir][monster.dir];

                    // sprite
                    int type = mapGroup.getType();
                    QString spriteName;
                    bool    flip = false;

                    if (tableDir == eMonsterDirDown)
                    {
                        // facing
                        if (monster.displayAttackTimer.isRunning())
                        {
                            // attack
                            spriteName = monstersDatas[type].attackImage;
                            if (spriteName.isEmpty() == true)
                                spriteName = monstersDatas[type].frontImage;
                        }
                        else
                        {
                            // normal
                            spriteName = monstersDatas[type].frontImage;
                        }
                    }
                    else if (tableDir == eMonsterDirUp)
                    {
                        // back
                        spriteName = monstersDatas[type].backImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                    }
                    else
                    {
                        // side
                        spriteName = monstersDatas[type].sideImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                        else if (tableDir == eMonsterDirLeft)
                            flip = true;
                    }

                    if (monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1] == (float)row / 2.0)
                    {
                        // table pos
                        float tablePosX = tablePos.x * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][0];
                        float tablePosY = tablePos.y * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1];

                        // screen pos
                        QImage  sprite = fileCache.getImage(spriteName.toStdString());
                        CVec2   pos = getMonsterPos(tablePosX, tablePosY);

                        // compute the scale based on the reference position (the nearest)
                        CVec2   pos0, pos1;
                        float   scale;

                        pos0 = getMonsterPos(WALL_TABLE_WIDTH, (WALL_TABLE_HEIGHT - 1) * 2 + 1);
                        pos1 = getMonsterPos(WALL_TABLE_WIDTH, tablePosY);
                        scale = 2.37 * (float)(pos1.x - 112) / (float)(pos0.x - 112);

                        // scale the monster in a temporary image
                        QImage  scaledMonster(sprite.size() * scale, QImage::Format_ARGB32);
                        scaledMonster.fill(TRANSPARENT);
                        graph2D.drawImageScaled(&scaledMonster, CVec2(), sprite, scale);

                        // darken the monster based on it's distance
                        float shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePosY - 1) * 0.13f;
                        graph2D.darken(&scaledMonster, shadow);

                        // draw the monster (pos relative to its feets)
                        pos.x -= scaledMonster.width() / 2;
                        pos.y -= scaledMonster.height() - 9 * scale;
                        graph2D.drawImage(image, pos, scaledMonster, 0, flip, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));
                    }
                }
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsters::update()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2& group = monsterGroups[i];

        group.update(mapGroup.mPos, mapGroup.getType());
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::update(CMonsterGroup2& group, CVec2 mapPos, uint8_t type, bool isInFront)
{
    displayAttackTimer.update();

    switch (group.state)
    {
        case eMonsterState_Wandering:
            wander();
            break;

        case eMonsterState_Fight:
            fight(mapPos, type, isInFront);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::wander()
{

}

//---------------------------------------------------------------------------------------------
void CMonster::initNextAttack(uint8_t type)
{
    int minTime = monsters.monstersDatas[type].timeBetweenAttack;
    int nextTime = minTime + RANDOM(40) - 10;

    if (minTime > 150)
        nextTime += RANDOM(80) - 20;

    nextAttackTimer.set(nextTime, true);
}

//---------------------------------------------------------------------------------------------
// choose the player this monster will attack
uint8_t CMonster::chooseTarget(CVec2 mapPos)
{
    int rowToSearch;

    // check if there is at least one living champion in the front row
    rowToSearch = 1;

    for (int i = 0; i < 4; ++i)
    {
        if (interface.isChampionEmpty(i) == false &&
            interface.isChampionDead(i) == false &&
            CMonsterGroup2::getChampionRow(mapPos, i) == 0)
        {
            rowToSearch = 0;
            break;
        }
    }

    // list the possible targets
    uint8_t possibleTargets[2];
    int nbTargets = 0;

    for (int i = 0; i < 4; ++i)
    {
        if (interface.isChampionEmpty(i) == false &&
            interface.isChampionDead(i) == false &&
            CMonsterGroup2::getChampionRow(mapPos, i) == rowToSearch)
        {
            possibleTargets[nbTargets++] = i;
        }
    }

    // choose a target
    return possibleTargets[RANDOM(nbTargets)];
}

//---------------------------------------------------------------------------------------------
// TODO: luck, experience, champ max damage, wounds, poison, parry
void CMonster::fight(CVec2 mapPos, uint8_t type, bool isInFront)
{
    if (nextAttackTimer.update() == true)
    {
        // if the monster is in the front line of the group
        if (isInFront == true)
        {
            // play attack sound
            QString soundName = monsters.monstersDatas[type].attackSound;

            if (soundName.isEmpty() == false)
                sound->play(mapPos, soundName.toLocal8Bit().constData());

            // show monster's animation
            int attackTime = monsters.monstersDatas[type].attackDisplayTime;
            displayAttackTimer.set(attackTime, true);

            // choose a target champion
            int target = chooseTarget(mapPos);

            // compute monster's and champion's dexterities
            int monsterDexterity = monsters.monstersDatas[type].dexterity + game.levelMultiplier() * 2 + RANDOM(32) - 16;
            int championDexterity = game.characters[target].getDexterity();

            // check if we hit
            if (interface.isSleeping() == true ||
                monsterDexterity > championDexterity ||
                RANDOM(4) == 0)
            {
                // compute damages (here there should be a chance to parry)
                int damages = monsters.monstersDatas[type].attack + game.levelMultiplier() * 2 + RANDOM(16);

                // randomize damages (quite a complicated way...)
                damages /= 2;
                damages += RANDOM(damages) + RANDOM(4);
                damages += RANDOM(damages);
                damages /= 4;
                damages += RANDOM(4) + 1;

                if (RANDOM(2))
                    damages -= RANDOM(damages / 2 + 1) - 1;

                // hit the champion
                static char soundName[256];
                sprintf(soundName, "sound/Champion_Damaged_%d.wav", RANDOM(4) + 1);
                sound->play(player.pos, soundName);
                game.characters[target].hitChampion(target, damages);
            }

            // if we are sleeping, wake up
            if (interface.isSleeping() == true)
                interface.wakeUp();
        }
        initNextAttack(type);
    }
}

//---------------------------------------------------------------------------------------------
CMonsterGroup2::CMonsterGroup2()
{
    state = eMonsterState_Wandering;
}

//---------------------------------------------------------------------------------------------
// returns the row of a monster depending on the player position (0 = front, 1 = center, 2 = back)
uint8_t CMonsterGroup2::getMonsterRow(CVec2 mapPos, EMonsterPos monsterPos)
{
    static const uint8_t monsterRow[4][5] =
    {
        // NW NE SW SE C
        {  0, 0, 2, 2, 1},   // up
        {  0, 2, 0, 2, 1},   // left
        {  2, 2, 0, 0, 1},   // down
        {  2, 0, 2, 0, 1}    // right
    };

    int playerPosFromMonster;

    if (player.pos.x < mapPos.x)
        playerPosFromMonster = 1; // left
    else if (player.pos.x > mapPos.x)
        playerPosFromMonster = 3; // right
    else if (player.pos.y < mapPos.y)
        playerPosFromMonster = 0; // up
    else
        playerPosFromMonster = 2; // down

    return monsterRow[playerPosFromMonster][monsterPos - eMonsterPosNorthWest];
}

//---------------------------------------------------------------------------------------------
// returns the row of champions depending on the monsters group position (0 = front, 1 = back)
uint8_t CMonsterGroup2::getChampionRow(CVec2 mapPos, int championNum)
{
    static const uint8_t championRow[4][4] =
    {
        // 1  2  3  4
        {  0, 0, 1, 1},   // up
        {  0, 1, 0, 1},   // left
        {  1, 1, 0, 0},   // down
        {  1, 0, 1, 0}    // right
    };

    CVec2   localMonsterPos = player.getLocalFromPos(mapPos);
    int monsterPosFromPlayer;

    if (localMonsterPos.x < 0)
        monsterPosFromPlayer = 1; // left
    else if (localMonsterPos.x > 0)
        monsterPosFromPlayer = 3; // right
    else if (localMonsterPos.y < 0)
        monsterPosFromPlayer = 0; // up
    else
        monsterPosFromPlayer = 2; // down

    return championRow[monsterPosFromPlayer][championNum];
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::update(CVec2 mapPos, uint8_t type)
{
    // check if there is at least one living monster in the front row of the group (or at the center)
    bool    isFrontRowRowAlive = false;

    for (int i = 0; i < 4; ++i)
    {
        if (monsters[i].pos != eMonsterPosNone &&
            getMonsterRow(mapPos, monsters[i].pos) < 2)
        {
            isFrontRowRowAlive = true;
            break;
        }
    }

    // update all the monsters of the group
    for (int i = 0; i < 4; ++i)
        if (monsters[i].pos != eMonsterPosNone)
        {
            uint8_t row = getMonsterRow(mapPos, monsters[i].pos);
            bool    isInFront = (isFrontRowRowAlive == true && row < 2) ||
                                (isFrontRowRowAlive == false && row == 2);
            monsters[i].update(*this, mapPos, type, isInFront);
        }

    // handle group behavior for each state
    switch (state)
    {
        case eMonsterState_Wandering:
            wander(mapPos, type);
            break;

        case eMonsterState_Fight:
            fight(mapPos, type);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::wander(CVec2 mapPos, uint8_t type)
{
    // transition to fight state ?
    CVec2   dist = player.pos - mapPos;
    bool    isNearPlayer = (dist.x == 0 && ABS(dist.y) == 1) || (dist.y == 0 && ABS(dist.x) == 1);

    if (isNearPlayer == true)
    {
        state = eMonsterState_Fight;

        // direction towards the player
        EMonsterDir dir;

        if (dist.x > 0)
            dir = eMonsterDirRight;
        else if (dist.x < 0)
            dir = eMonsterDirLeft;
        else if (dist.y > 0)
            dir = eMonsterDirDown;
        else
            dir = eMonsterDirUp;

        // initialize each monster of the group
        for (int i = 0; i < 4; ++i)
            if (monsters[i].pos != eMonsterPosNone)
            {
                monsters[i].dir = dir;
                monsters[i].initNextAttack(type);
            }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::fight(CVec2 mapPos, uint8_t type)
{
    // transition to wandering state ?
    CVec2   dist = player.pos - mapPos;
    bool    isNearPlayer = (dist.x == 0 && ABS(dist.y) == 1) || (dist.y == 0 && ABS(dist.x) == 1);

    if (isNearPlayer == false)
    {
        state = eMonsterState_Wandering;
    }
}
